"""
Define Qt key and modifier constants for compatibility between Qt5 and Qt6.
"""

from matplotlib.backends.qt_compat import QtCore

try:
    assert hasattr(QtCore.Qt, "Key")  # Qt6
    Key_Question = QtCore.Qt.Key.Key_Question
    Key_Q = QtCore.Qt.Key.Key_Q
    Key_P = QtCore.Qt.Key.Key_P
    Key_H = QtCore.Qt.Key.Key_H
    Key_Z = QtCore.Qt.Key.Key_Z
    Key_C = QtCore.Qt.Key.Key_C
    Key_V = QtCore.Qt.Key.Key_V
    Key_S = QtCore.Qt.Key.Key_S
    Key_K = QtCore.Qt.Key.Key_K
    Key_Space = QtCore.Qt.Key.Key_Space
    Key_Home = QtCore.Qt.Key.Key_Home
    Key_End = QtCore.Qt.Key.Key_End
    Key_MediaPrevious = QtCore.Qt.Key.Key_MediaPrevious
    Key_MediaNext = QtCore.Qt.Key.Key_MediaNext
    Key_MediaPlay = QtCore.Qt.Key.Key_MediaPlay
    Key_MediaPause = QtCore.Qt.Key.Key_MediaPause
    Key_MediaTogglePlayPause = QtCore.Qt.Key.Key_MediaTogglePlayPause
    Key_Left = QtCore.Qt.Key.Key_Left
    Key_Right = QtCore.Qt.Key.Key_Right

    NoModifier = QtCore.Qt.KeyboardModifier.NoModifier
    ControlModifier = QtCore.Qt.KeyboardModifier.ControlModifier
    ShiftModifier = QtCore.Qt.KeyboardModifier.ShiftModifier
except AttributeError:  # Qt5
    Key_Question = QtCore.Qt.Key_Question
    Key_Q = QtCore.Qt.Key_Q
    Key_P = QtCore.Qt.Key_P
    Key_H = QtCore.Qt.Key_H
    Key_Z = QtCore.Qt.Key_Z
    Key_C = QtCore.Qt.Key_C
    Key_V = QtCore.Qt.Key_V
    Key_S = QtCore.Qt.Key_S
    Key_K = QtCore.Qt.Key_K
    Key_Space = QtCore.Qt.Key_Space
    Key_Home = QtCore.Qt.Key_Home
    Key_End = QtCore.Qt.Key_End
    Key_MediaPrevious = QtCore.Qt.Key_MediaPrevious
    Key_MediaNext = QtCore.Qt.Key_MediaNext
    Key_MediaPlay = QtCore.Qt.Key_MediaPlay
    Key_MediaPause = QtCore.Qt.Key_MediaPause
    Key_MediaTogglePlayPause = QtCore.Qt.Key_MediaTogglePlayPause
    Key_Left = QtCore.Qt.Key_Left
    Key_Right = QtCore.Qt.Key_Right

    NoModifier = QtCore.Qt.NoModifier
    ControlModifier = QtCore.Qt.ControlModifier
    ShiftModifier = QtCore.Qt.ShiftModifier



"""
abracatabra
===========
A package for creating tabbed plot windows with matplotlib in a Qt environment.

Windows can be created with one or more tab groups. Each tab group can contain
one or more tabs, where is tab is a matplotlib Figure. The library provides
functions to show all open windows, update them, and a fun `abracatabra`
function to display all windows with a magical touch.

This package provides:
- `TabbedPlotWindow`: The main class for creating and managing tabbed plot windows.
- `show_all_windows`: Displays all open tabbed plot windows.
- `update_all_windows`: Updates all open tabbed plot windows.
- `animate_all_windows`: Animates all open tabbed plot windows based on
    registered callbacks.
- `abracatabra`: A fun function to display all open tabbed plot windows.
- `is_interactive`: Checks if the current environment is interactive
    (e.g., IPython or Jupyter).
- `__version__`: The version of the abracatabra package.
"""

from pathlib import Path

from .tabbed_plot_window import TabbedPlotWindow, is_interactive
from .__about__ import __version__


def show_all_windows(tight_layout: bool = False, block: bool | None = None) -> None:
    """
    Shows all created windows.

    Args:
        tight_layout: If True, applies a tight layout to all figures.
        block: If True, block and run the GUI until all windows are
            closed, either individually or by pressing <ctrl+c> in the terminal.
            If False, the function returns immediately after showing the windows
            and you are responsible for ensuring the GUI event loop is running
            (interactive environments do this for you).
            Defaults to False in interactive environments, otherwise True.
    See Also
    -----
    `abracatabra()`: shows all created windows with a touch of magic!
    `update_all_windows()`: updates all open tabbed plot windows.
    `is_interactive()`: checks if the current environment is interactive.
    """
    TabbedPlotWindow.show_all(tight_layout, block)


def update_all_windows(delay_seconds: float = 0.0, callback_idx: int = 0) -> float:
    """
    Updates all open tabbed plot windows. This is similar to pyplot.pause()
    and is generally used to update the figure in a loop, e.g., an animation.
    This function only updates active tabs in each window, so inactive tabs are
    skipped to save time.

    Args:
        delay_seconds: The minimum delay in seconds before returning. If
            windows are updated faster than this, this function will block until
            `delay_seconds` seconds have passed. If the windows take longer than
            `delay_seconds` seconds to update, the function execution time will
            be greater than `delay_seconds`.
        callback_idx: An index passed to registered animation callbacks to
            specify which frame to draw.

    Returns:
        update_time: The amount of time (seconds) taken to update the windows.
    See Also
    -----
    `show_all_windows()`: shows all created windows.
    `animate_all_windows()`: animates all open tabbed plot windows based on
        registered callbacks.
    `abracatabra()`: shows all created windows with a touch of magic!
    """
    return TabbedPlotWindow.update_all(delay_seconds, callback_idx)


def animate_all_windows(
    frames: int,
    ts: float,
    step: int = 1,
    speed_scale: float = 1.0,
    print_timing: bool = False,
    use_player: bool = False,
    hold: bool = True,
) -> None:
    """
    Animates all created windows by repeatedly calling `update_all_windows()` in
    a loop for the given number of frames. This is a convenience function for
    simple animations. For this to work well, you should have already
    registered animation callbacks for each tab that will be animated.

    Args:
        frames: The number of frames to animate.
        ts (seconds): The time step between frames in seconds. The intention
            is that this time step matches real time, e.g., a simulation
            that saves data every `ts` seconds.
        step: The step size between frames. For example, if you want to animate
            every 2nd frame, set step=2. This is useful if your animation is
            running slower than real time and you want to draw batches of data
            between a single frame.
        speed_scale: A scaling factor for the speed of the animation.
            For example, if speed_scale=2.0, the animation will run twice as
            fast (i.e., half the time step between frames), meaning that a
            10 sec simulation should take 5 sec to animate.
        print_timing: If True, prints timing information for each frame, including
            the running animation time and wall time. Also prints hints after the
            animation is done on how to improve performance if the animation is
            running slower than real time.
        use_player: Specifies whether to use an animation player window with media
            controls (play, pause, step, etc.) to control the animation. If an
            animation player has already been added to a tab, it will be used
            (even if `use_player` is False); otherwise, a new animation player
            window will be created.
        hold: Specify whether to keep the windows open (blocking code) at the last
            frame when the animation is complete. Essentially whether to call
            `show_all_windows()` at the end or not.
    See Also
    -----
    `update_all_windows()`: updates all open tabbed plot windows.
    """
    TabbedPlotWindow.animate_all(
        frames, ts, step, speed_scale, print_timing, use_player, hold
    )


def save_animations(frames: int, ts: float, **kwargs) -> None:
    """
    Opens a file dialog that will display all tabs with registered animation
    callbacks and allow you to save the animation for each tab as a video file.
    This function uses `matplotlib.animation.FuncAnimation` under the hood, so
    it may not be the most efficient way to save animations, but it is a convenient way to save animations without having to manually create `FuncAnimation` objects for each tab. Note that this function will save the animation for all tabs with registered animation callbacks, so if you only want to save the animation for a specific tab, you should use `FuncAnimation` directly on that tab's figure.
    """
    TabbedPlotWindow.save_animations(frames, ts, **kwargs)


def set_animation_savedir(save_dir: str | Path | None) -> None:
    """
    Sets the default directory to save animations when using the animation
    player or the `save_animations` dialog.

    Args:
        save_dir: The directory to save the videos in. If None, will reset
            to the default save directory, which looks for the first valid
            path in this order: 1) "~/Videos" 2) "~/Documents" 3) "~".
    """
    TabbedPlotWindow.set_animation_savedir(save_dir)


def abracatabra(
    tight_layout: bool = False, block: bool | None = None, verbose: bool = True
) -> None:
    """
    A more fun equivalent to `show_all_windows()`. Shows all created windows.

    Args:
        tight_layout: If True, applies a tight layout to all figures.
        block: If True, block and run the GUI until all windows are closed, either
            individually or by pressing <ctrl+c> in the terminal. If False, the
            function returns immediately after showing the windows and you are
            responsible for ensuring the GUI event loop is running (interactive
            environments do this for you).  Defaults to False in interactive
            environments, otherwise True.
        verbose: If True, prints a message when showing windows.

    See Also
    -----
    `show_all_windows()` : shows all created windows.
    `update_all_windows()` : updates all open tabbed plot windows.
    `is_interactive()` : checks if the current environment is interactive.
    """
    if verbose:
        print("Abracatabra! 🪄✨")
    for window in TabbedPlotWindow._registry.values():
        window.qt.setWindowIcon(window._icon2)
    TabbedPlotWindow.show_all(tight_layout, block)


del Path

__all__ = [
    "TabbedPlotWindow",
    "show_all_windows",
    "update_all_windows",
    "animate_all_windows",
    "abracatabra",
    "is_interactive",
    "__version__",
]



# system imports
import signal
import sys
import os
import time
import random
from typing import Callable
from pathlib import Path

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

from matplotlib.figure import Figure
from matplotlib.backends.qt_compat import QtWidgets, QtCore, QtGui

# from PySide6 import QtWidgets, QtCore, QtGui

# Fix plot font types to work in paper sumbissions (Don't use type 3 fonts)
import matplotlib

matplotlib.rcParams["pdf.fonttype"] = 42
matplotlib.rcParams["ps.fonttype"] = 42

from .animation_player import AnimationPlayer
from .figure_widget import FigureWidget
from .tabbed_figure_widget import TabbedFigureWidget
from .tab_group_container import TabGroupContainer
from . import keys

# if sys.modules.get('IPython') is not None:
try:
    from IPython.core.getipython import get_ipython

    _ipython = get_ipython()
except ImportError:
    _ipython = None

if _ipython is not None:
    from IPython.utils.capture import capture_output

    with capture_output() as captured:  # suppress output
        # register IPython event loop to Qt - prevents need to call app.exec()
        _ipython.run_line_magic("gui", "qt")

    # SIGINT handles ctrl+c. The following lines allow it to kill without errors.
    # Using sys.exit(0) in IPython stops script execution, but not the kernel.
    signal.signal(signal.SIGINT, lambda sig, frame: sys.exit(0))
else:
    # Use SIG_DFL (default) rather than letting Qt handle ctrl+c.
    # Qt throws a KeyboardInterrupt exception, but only when the mouse hovers
    # over the window or some other Qt action causes events to process.
    signal.signal(signal.SIGINT, signal.SIG_DFL)


def is_interactive() -> bool:
    """
    Check if the current environment is interactive (e.g., IPython or Jupyter).

    Returns:
        out (bool): True if the environment is interactive, False otherwise.
    """
    return bool(_ipython)


icon_dir = os.path.join(os.path.dirname(__file__), "icons")
# icon_files = ['tabplot.svg'] + [f'abracatabra{i}.svg' for i in range(1, 4)]
icon_files = ["tabplot.svg", f"abracatabra{random.choice([1, 2, 3])}.svg"]
icon_paths = [os.path.join(icon_dir, icon) for icon in icon_files]


class TabbedPlotWindow:
    """
    A class to create a tabbed plot window where the tabs are matplotlib
    figures. The window can also be divided into multiple tab groups, each
    containing multiple tabs.

    Attributes:
        `id`: Unique identifier for the window.
        `qt`: The underlying Qt MainWindow object.
        `tab_groups`: A container for the tab groups in the window. Can be
            accessed like a 2D array, e.g. `window.tab_groups[0, 0]` for the
            tab group in the first row and first column.
    Methods:
        `add_figure_tab`: Method to add a new figure tab to the window.
        `add_custom_tab`: Method to add a new custom widget tab to the window.
        `register_animation_callback`: Method to register a callback function for
            how to update the figure or custom widget in a tab.
        `update`: Method to update the figure on the active tab.
        `get_keyboard_shortcuts_str`: Returns a string with the keyboard shortcuts
            for the window.
        `display_keyboard_shortcuts`: Displays a message box with the keyboard
            shortcuts for the window.
        `set_size`: Method to set the size of the window in either pixels or a
            percentage of the screen.
        `apply_tight_layout`: Applies a tight layout to the figure in each tab.
        `enable_tab_autohide`: Enables auto-hiding of tabs in the window.
        `set_tab_position`: Sets the position of the tab bar in the window.
        `set_tab_fontsize`: Sets the font size of the tab labels in the window.
    Static Methods:
        `show_all`: Shows all created windows.
        `update_all`: Updates all created windows.
        `animate_all`: Animates all created windows.
        `close_all_windows`: Closes all created windows.
        `get_screen_size`: Returns the size of the screen in pixels.
    """

    _app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv)
    _registry: dict[str, Self] = {}
    _latest_id = None
    _count = 0
    # _icons = [QtGui.QIcon(icon) for icon in icon_paths]
    _icon1 = QtGui.QIcon(icon_paths[0])
    _icon2 = QtGui.QIcon(icon_paths[1])

    def __new__(
        cls,
        window_id: str | int | None = None,
        nrows: int | list[int] = 1,
        ncols: int | list[int] = 1,
        size: tuple[int | float, int | float] = (0.6, 0.8),
        open_window: bool = True,
        add_animation_player: bool = False,
        autohide_tabs: bool = False,
        tab_position: str = "top",
        tab_fontsize: int = 8,
    ) -> Self:
        if window_id is None:
            # Generate a unique identifier if none is provided
            id_ = str(len(cls._registry) + 1)
            while id_ in cls._registry:
                id_ = str(int(id_) + 1)
        else:
            id_ = str(window_id)

        # Return instance if it exists
        if id_ in cls._registry:
            return cls._registry[id_]

        # Create a new instance if it does not exist
        instance = super().__new__(cls)
        cls._registry[id_] = instance
        cls._latest_id = id_
        cls._count += 1
        return instance

    def __init__(
        self,
        window_id: str | int | None = None,
        nrows: int | list[int] = 1,
        ncols: int | list[int] = 1,
        size: tuple[int | float, int | float] = (0.6, 0.8),
        open_window: bool = True,
        add_animation_player: bool = False,
        autohide_tabs: bool = False,
        tab_position: str = "top",
        tab_fontsize: int = 8,
    ):
        """
        Creates a new tabbed plot window with the given ID and size. If a window
        with the same ID already exists, it will return that instance instead of
        creating a new one. Only nrows or ncols can be a list, not both.

        Args:
            window_id: The ID of the window. If None, a unique ID will be created
                based on the number of existing windows.
            nrows: The number of rows of tab groups. If a list, specifies the
                number of columns in each row, e.g. nrows=[1,2] would have 1 row
                in the first column and 2 rows in the second column.
            ncols: The number of columns of tab groups. If a list, specifies the
                number of rows in each column, e.g. ncols=[1,2] would have 1
                column in the first row and 2 columns in the second row.
            size: Size of the window (width, height). If an int, the value will
                be treated as pixels. If a float, it will be treated as a
                percentage of the screen size (based on your PRIMARY DISPLAY).
            open_window: If True, the window will be displayed immediately after
                creation. Otherwise, it will be hidden until another method is
                called to show it.
            autohide_tabs: If True, the tab bar will auto-hide when there is only
                one tab in a tab group.
            tab_position: The position of the tab bar. This can be 'top', 'bottom',
                'left', or 'right' as well as 'north', 'south', 'east', or 'west'
                (logic only looks at first character).
            tab_fontsize: The font size of the tab labels.
        """
        if hasattr(self, "id"):
            return
        # super().__init__()
        self.qt = QtWidgets.QMainWindow()
        self.id = str(self._latest_id)
        self.qt.setWindowTitle(f"Plot Window: {self.id}")
        self.set_size(size)
        self.qt.setWindowIcon(TabbedPlotWindow._icon1)
        central_widget = QtWidgets.QWidget()
        self.qt.setCentralWidget(central_widget)
        self._setup_shortcuts()
        # self.qt.keyPressEvent = self._key_press_event

        if add_animation_player:
            central_layout = QtWidgets.QVBoxLayout(central_widget)
            central_layout.setSpacing(0)
            central_layout.setContentsMargins(0, 0, 0, 0)

            main_widget = QtWidgets.QWidget()
            central_layout.addWidget(main_widget)
            player = AnimationPlayer(parent=self.qt)
            # player.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
            central_layout.addWidget(player, stretch=0)
        else:
            main_widget = central_widget

        row_major = True
        tab_groups = []
        if isinstance(nrows, int) and isinstance(ncols, int):
            ## create a grid layout with nrows x ncols
            main_layout = QtWidgets.QGridLayout(main_widget)
            if nrows < 1 or ncols < 1:
                raise ValueError(
                    f"Can not have {nrows} rows or {ncols} columns. Must be at least 1."
                )
            for r in range(nrows):
                row: list[TabbedFigureWidget] = []
                for c in range(ncols):
                    main_layout.setColumnStretch(c, 1)
                    widget = TabbedFigureWidget(
                        autohide_tabs, tab_position, tab_fontsize
                    )
                    # widget.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
                    row.append(widget)
                    main_layout.addWidget(widget, r, c)
                tab_groups.append(row)
        elif isinstance(ncols, list):
            ## create a vertical layout nested with horizontal layouts
            if isinstance(nrows, list):
                raise ValueError("Either nrows or ncols can be a list, not both.")
            main_layout = QtWidgets.QVBoxLayout(main_widget)
            for r in ncols:
                if r < 1:
                    raise ValueError(f"Can not have {r} columns. Must be at least 1.")
                widget_row = QtWidgets.QWidget()
                hlayout = QtWidgets.QHBoxLayout(widget_row)
                hlayout.setContentsMargins(0, 0, 0, 0)
                row = []
                for c in range(r):
                    widget = TabbedFigureWidget(
                        autohide_tabs, tab_position, tab_fontsize
                    )
                    # widget.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
                    row.append(widget)
                    hlayout.addWidget(widget)
                tab_groups.append(row)
                main_layout.addWidget(widget_row)
        elif isinstance(nrows, list):
            ## create a horizontal layout nested with vertical layouts
            if isinstance(ncols, list):
                raise ValueError("Either nrows or ncols can be a list, not both.")
            row_major = False
            main_layout = QtWidgets.QHBoxLayout(main_widget)
            for c in nrows:
                if c < 1:
                    raise ValueError(f"Can not have {c} columns. Must be at least 1.")
                widget_col = QtWidgets.QWidget()
                vlayout = QtWidgets.QVBoxLayout(widget_col)
                vlayout.setContentsMargins(0, 0, 0, 0)
                col = []
                for r in range(c):
                    widget = TabbedFigureWidget(
                        autohide_tabs, tab_position, tab_fontsize
                    )
                    # widget.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
                    col.append(widget)
                    vlayout.addWidget(widget)
                tab_groups.append(col)
                main_layout.addWidget(widget_col)
        else:
            raise ValueError("Invalid values for `nrows` and `ncols`")
        main_layout.setContentsMargins(0, 0, 0, 0)
        self.tab_groups = TabGroupContainer(tab_groups, row_major)

        # Register close event handler
        self.qt.closeEvent = self.close_event

        if open_window:
            self.qt.show()

    def add_figure_tab(
        self,
        tab_id: str,
        blit: bool = False,
        include_toolbar: bool = True,
        add_animation_player: bool = False,
        row: int = 0,
        col: int = 0,
    ) -> Figure:
        """
        Adds a new tab to the window with the given ID and returns the Figure
        created for that tab. If a tab with the same ID already exists, the
        existing Figure will be returned instead of creating a new one.

        Args:
            tab_id: The ID of the tab.
            blit: Whether blitting will be used with the Figure in this tab. If
                True, you are responsible for managing the background and
                updating individual artists.
            include_toolbar: Whether to display a matplotlib toolbar with the
                Figure in this tab.
            add_animation_player: Whether to include an animation player widget
                in this tab (play, pause, etc.). Only works if animation
                callbacks are registered.
            row: The row index of the tab group to add the tab to.
            col: The column index of the tab group to add the tab to.

        Returns:
            figure: The matplotlib figure in this tab.

        Notes
        -----
        Can alternatively call `add_figure_tab` directly on a tab group:
        ```python
        window = TabbedPlotWindow(ncols=2)
        fig_left = window.add_figure_tab("left", row=0, col=0)
        fig_right = window.tab_groups[0, 1].add_figure_tab("right")
        ```
        """
        fig_tab_widget = self.tab_groups[row, col]
        figure = fig_tab_widget.add_figure_tab(
            tab_id, blit, include_toolbar, add_animation_player
        )
        return figure

    def add_custom_tab(
        self,
        widget: QtWidgets.QWidget,
        tab_id: str | int,
        add_animation_player: bool = False,
        row: int = 0,
        col: int = 0,
    ) -> None:
        """
        Adds a new tab, containing the provided widget, to the window with the
        given ID as the tab label. The ID must be unique within the tab group.

        Args:
            widget: The custom Qt widget to add as a tab.
            tab_id: The ID of the tab.
            add_animation_player: Whether to include an animation player widget
                in this tab (play, pause, etc.). Only works if animation
                callbacks are registered.
            row: The row index of the tab group to add the tab to.
            col: The column index of the tab group to add the tab to.

        Notes
        -----
        Can alternatively call `add_custom_tab` directly on a tab group:
        ```python
        window = TabbedPlotWindow(ncols=2)
        tab_left = window.add_custom_tab(widget, "left", row=0, col=0)
        tab_right = window.tab_groups[0, 1].add_custom_tab(widget, "right")
        ```
        """
        tab_widget = self.tab_groups[row, col]
        tab_widget.add_custom_tab(widget, tab_id, add_animation_player)
        return

    def register_animation_callback(
        self, callback: Callable[[int], None], tab_id: str, row: int = 0, col: int = 0
    ) -> None:
        """
        Registers a callback function for how to update the figure or the custom
        widget in the specified tab during an animation.

        Args:
            callback: A function specifying how to update the widget. The function
                should take a single integer argument, which is the index of the
                current frame in the animation to draw. Registering callbacks
                allows abracatabra to better manage the timing of updates.
            tab_id: The ID/title of the tab.
            row: The row index of the tab group containing the tab.
            col: The column index of the tab group containing the tab.
        """
        tab_widget = self.tab_groups[row, col][tab_id]
        tab_widget.register_animation_callback(callback)
        return

    def show(self) -> None:
        """
        Shows the window if it is not already visible.
        """
        self.qt.show()

    def update(self, callback_idx: int = 0) -> None:
        """
        This will update the figure on the active (visible) tabs. Similar to
        pyplot.pause(), but for the current tab on this window. No additional
        time delay is added to the function, so it will return immediately after
        updating the figure.
        """
        if not self.qt.isVisible():
            self.qt.show()
        for tabs in self.tab_groups:
            tabs.update_active_tab(callback_idx)

    def _key_press_event(self, event: QtGui.QKeyEvent):
        """
        Qt event function - DO NOT CALL DIRECTLY.

        This method is called when a key is pressed while the window is in
        focus. It will close the window if 'q' is pressed, or close all windows
        if 'ctrl+q' is pressed.
        """
        player = AnimationPlayer.instance()
        print("window key press:", event.text())
        # if player and player._parent is None:
        #     TabbedPlotWindow._app.postEvent(player, event)
        #     # TabbedPlotWindow._app.sendEvent(player, event)

        # key_used = True
        # match event.key():
        #     case keys.Key_Question:
        #         self.display_keyboard_shortcuts()
        #     case keys.Key_Q:
        #         if event.modifiers() & keys.ControlModifier:
        #             TabbedPlotWindow.close_all_windows()
        #             if player:
        #                 player.close()
        #         else:
        #             self.qt.close()
        #     case _:
        #         key_used = False
        #
        # if not key_used and player and player._parent is None:
        #     player.keyPressEvent(event)
        # else:
        #     QtWidgets.QMainWindow.keyPressEvent(self.qt, event)

    def _setup_shortcuts(self):
        """
        Sets up keyboard shortcuts for the window.
        """
        shortcuts = {
            keys.Key_Question: self.display_keyboard_shortcuts,
            keys.Key_Q: self.qt.close,
            keys.Key_Q | keys.ControlModifier: TabbedPlotWindow.close_all_windows,
        }
        for key, func in shortcuts.items():
            shortcut = QtGui.QShortcut(QtGui.QKeySequence(key), self.qt)
            shortcut.setContext(QtCore.Qt.ShortcutContext.WindowShortcut)
            shortcut.activated.connect(func)

    def get_keyboard_shortcuts_str(self) -> str:
        help = (
            "General Controls:\n"
            "    q: Close focused window\n"
            "    Ctrl+q: Close all windows\n"
            "    Tab: Focus next tab group\n"
            "    Shift+Tab: Focus previous tab group\n"
            "    ?: Show this help dialog\n\n"
            "Tab Widget Controls:\n"
            "    Ctrl+Tab: Next tab\n"
            "    Ctrl+Shift+Tab: Previous tab\n"
            f"\n{FigureWidget.help_text}"
        )
        player = AnimationPlayer.instance()
        if player is not None:
            help += f"\n{player.help_text}"
        return help

    def display_keyboard_shortcuts(self) -> None:
        """
        Displays a message box with the keyboard shortcuts for the window.
        """
        title = "Plot Window Keyboard Shortcuts"
        help = self.get_keyboard_shortcuts_str()
        QtWidgets.QMessageBox.information(self.qt, title, help)

    def close_event(self, event: QtGui.QCloseEvent) -> None:
        """
        Qt event function - DO NOT CALL DIRECTLY.

        This method is called when the window is closed. It will remove the
        window from the list of windows and check if there are any other windows
        open. If not, it will exit the application.
        """
        event.accept()
        del TabbedPlotWindow._registry[self.id]
        TabbedPlotWindow._count -= 1
        if TabbedPlotWindow._count != 0:
            return
        player = AnimationPlayer.instance()
        if player:
            player.close()

    def set_size(self, size: tuple[int | float, int | float]) -> None:
        """
        Sets the size of the window in either pixels or a percentage of the
        screen.

        Args:
            size: Size of the window (width, height). If an int, the value will
                be treated as pixels. If a float, it will be treated as a
                percentage of the screen size.
        """
        width, height = size
        if width <= 0 or height <= 0:
            raise ValueError(f"Invalid size: {size}. Values must be positive.")
        screen_size = self.qt.screen().size()
        screen_width, screen_height = screen_size.width(), screen_size.height()
        if isinstance(width, float):
            if width > 1.0:
                raise ValueError(f"Width percentage {width} must be between 0 and 1.")
            width = int(screen_width * width)
        if isinstance(height, float):
            if height > 1.0:
                raise ValueError(f"Height percentage {height} must be between 0 and 1.")
            height = int(screen_height * height)
        width = min(width, screen_width)
        height = min(height, screen_height)
        self.qt.resize(width, height)

    def apply_tight_layout(self):
        """
        Applies a tight layout to the figure in each tab of the window, even if
        they are not the active Figure. Same as calling figure.tight_layout()
        directly on each Figure.
        """
        for tabs in self.tab_groups:
            current_index = tabs.currentIndex()
            for i in range(tabs.count()):
                tab = tabs.widget(i)
                if not isinstance(tab, FigureWidget):
                    continue
                tabs.setCurrentIndex(i)  # tab has to be active to apply tight layout
                tab.figure.tight_layout()
                tab.update_figure()
            tabs.setCurrentIndex(current_index)

    def enable_tab_autohide(self, enable: bool = True) -> None:
        """
        Enables auto-hiding of tabs in the window. This will hide the tab bar
        when there is only one tab in a tab group.

        Args:
            enable: Whether to enable auto-hiding of tabs.
        """
        for tabs in self.tab_groups:
            tabs.setTabBarAutoHide(enable)

    def set_tab_position(self, position: str) -> None:
        """
        Sets the position of the tab bar in the window.

        Args:
            position: The position of the tab bar. This can be 'top', 'bottom',
                'left', or 'right' as well as 'north', 'south', 'east', or 'west'
                (logic only looks at first character).
        """
        for tabs in self.tab_groups:
            tabs.set_tab_position(position)

    def set_tab_fontsize(self, fontsize: int) -> None:
        """
        Sets the font size of the tab labels in the window.

        Args:
            fontsize: The font size of the tab labels.
        """
        for tabs in self.tab_groups:
            tabs.set_tab_fontsize(fontsize)

    @staticmethod
    def show_all(tight_layout: bool = False, block: bool | None = None) -> None:
        """
        Shows all created windows.

        Args:
            tight_layout: If True, applies a tight layout to all figures.
            block: If True, block and run the GUI until all windows are closed,
                either individually or by pressing <ctrl+c> in the terminal. If
                False, the function returns immediately after showing the windows
                and you are responsible for ensuring the GUI event loop is running
                (interactive environments do this for you).  Defaults to False in
                interactive environments, otherwise True.
        """
        for key in list(TabbedPlotWindow._registry.keys()):
            if not key in TabbedPlotWindow._registry:
                continue  # in case window was closed during iteration
            window = TabbedPlotWindow._registry[key]
            if not window.qt.isVisible():
                window.qt.show()
            if tight_layout:
                window.apply_tight_layout()
        if block is None:
            block = not is_interactive()
        if not block:
            return
        if TabbedPlotWindow._count > 0 and TabbedPlotWindow._app is not None:
            try:
                TabbedPlotWindow._app.exec()  # type: ignore
            except:
                TabbedPlotWindow._app.exec_()  # for compatibility with Qt5

    @staticmethod
    def update_all(delay_seconds: float, callback_idx: int = 0) -> float:
        """
        Updates all created windows. This is similar to pyplot.pause() and is
        generally used to update the figure in a loop, e.g., an animation. This
        function only updates active tabs in each window, so inactive tabs are
        skipped to save time.

        Args:
            delay_seconds: The minimum delay in seconds before returning.
                If windows are updated faster than this, this function will
                block until `delay_seconds` seconds have passed. If the windows
                take longer than `delay_seconds` seconds to update, the function
                execution time will be greater than `delay_seconds`.
            callback_idx: An index passed to registered animation callbacks to
                specify which frame to draw.

        Returns:
            update_time: The amount of time (seconds) taken to update the windows.
        """
        start = time.perf_counter()
        for key in list(TabbedPlotWindow._registry.keys()):
            if not key in TabbedPlotWindow._registry:
                continue  # in case window was closed during iteration
            window = TabbedPlotWindow._registry[key]
            window.update(callback_idx)
        update_time = time.perf_counter() - start
        if TabbedPlotWindow._count > 0:
            remaining_delay = max(delay_seconds - update_time, 0.0)
            time.sleep(remaining_delay)
        return update_time

    @staticmethod
    def animate_all(
        frames: int,
        ts: float,
        step: int = 1,
        speed_scale: float = 1.0,
        print_timing: bool = False,
        use_player: bool = False,
        hold: bool = True,
    ) -> None:
        """
        Animates all created windows by repeatedly calling `update_all()` in a
        loop for the given number of frames. This is a convenience function for
        simple animations. For this to work well, you should have already
        registered animation callbacks for each tab that will be animated.

        Args:
            frames: The number of frames to animate.
            ts (seconds): The time step between frames in seconds. The intention
                is that this time step matches real time, e.g., a simulation
                that saves data every `ts` seconds.
            step: The step size between frames. For example, if you want
                to animate every 2nd frame, set step=2. This is useful if your
                animation is running slower than real time and you want to draw
                batches of data between a single frame.
            speed_scale: A scaling factor for the speed of the animation.
                For example, if speed_scale=2.0, the animation will run twice as
                fast (i.e., half the time step between frames), meaning that a
                10 sec simulation should take 5 sec to animate.
            print_timing: If True, prints timing information for each frame,
                including the running animation time and wall time. Also prints hints
                after the animation is done on how to improve performance if the
                animation is running slower than real time.
            use_player: Specifies whether to use an animation player window
                with media controls (play, pause, step, etc.) to control the
                animation. If an animation player has already been added to a tab,
                it will be used (even if `use_player` is False); otherwise, a new
                animation player window will be created.
            hold: Specify whether to keep the windows open (blocking code)
                at the last frame when the animation is complete. Essentially
                whether to call `show_all()` at the end or not.
        """
        if frames < 1 or step < 1:
            raise ValueError("Frames and step must be positive integers.")
        if ts <= 0 or speed_scale <= 0:
            raise ValueError("Time step and speed scale must be positive values.")
        if step / frames > 0.01:
            print("Warning: `step` is larger than 1% of `frames`.")

        delay = ts * step / speed_scale

        # Check if there are any windows to animate
        if TabbedPlotWindow._count == 0:
            print("No windows to animate. Skipping animation.")
            return

        if use_player:
            player = AnimationPlayer.instance() or AnimationPlayer()
            player.save_button.clicked.connect(
                lambda: TabbedPlotWindow.save_animations(frames, ts)
            )
            player.setFocus(QtCore.Qt.FocusReason.ActiveWindowFocusReason)

            for window in TabbedPlotWindow._registry.values():
                player.register_shortcuts(window.qt)

            def callback(frame: int):
                TabbedPlotWindow.update_all(0.0, frame)

            player.setup(frames, ts, step, callback)
            TabbedPlotWindow._app.processEvents()

            while player.isVisible() and TabbedPlotWindow._count > 0:
                start = time.perf_counter()
                stepped = player.step_frame()
                if not stepped:
                    TabbedPlotWindow._app.processEvents()
                update_time = time.perf_counter() - start
                if TabbedPlotWindow._count > 0:
                    remaining_delay = max(delay - update_time, 0.0)
                    time.sleep(remaining_delay)
            TabbedPlotWindow.close_all_windows()
            return

        start = time.perf_counter()
        for i in range(0, frames, step):
            TabbedPlotWindow.update_all(delay, i)

            if not print_timing:
                continue

            elapsed = time.perf_counter() - start
            print(
                f"animation time: {i*ts:.2f}s",
                f"real time: {elapsed:.2f}s",
                sep=" | ",
                end="\r",
            )
        # Ensure the final frame is drawn
        TabbedPlotWindow.update_all(0.0, frames - 1)

        if print_timing:
            print()  # newline after final frame printout

        real_time = time.perf_counter() - start
        sim_time = frames * ts
        actual_speed_scale = sim_time / real_time
        buffer_percent = 10.0
        percent_error = (speed_scale - actual_speed_scale) / speed_scale * 100.0
        if percent_error > buffer_percent:
            print("Your computer is not keeping up with the requested speeds!")
            print(f"Tried to run at {speed_scale:.1f}x speed,", end=" ")
            print(f"but actual speed was {actual_speed_scale:.1f}x.")
            if speed_scale > 1.0:
                print("Try decreasing 'speed_scale' or increasing 'step'")
            else:
                print("Try increasing 'step'")

        if hold:
            TabbedPlotWindow.show_all()
        TabbedPlotWindow.close_all_windows()

    @staticmethod
    def save_animations(
        frames: int, ts: float, save_dir: str | Path | None = None, **kwargs
    ) -> None:
        """
        Opens a dialog to save detected animations from all windows as video
        files. An animation is detected if a tab has a registered animation
        callback. If no animations are found, a message box will be displayed
        indicating that no animations were found. If animations are found, a
        dialog will be displayed with a list of the windows and tabs that have
        registered animation callbacks, along with a "Save" button for each one.

        Args:
            frames: The number of frames in the animation.
            ts (seconds): The time step between frames in seconds. This is used
                to set the frame rate of the saved video (frame rate = 1/ts).
            save_dir: The directory to save the videos in. If None, will reset
                to the default save directory (see `set_animation_savedir` method).
            **kwargs: Additional keyword arguments to pass to the `save_animation`
                method of each `FigureWidget`. This can include arguments for the
                codec, bitrate, etc. See `matplotlib.animation.FuncAnimation`
                documentation for more details on available arguments.
        """
        animations: dict[TabbedPlotWindow, list[FigureWidget]] = {}
        for key in list(TabbedPlotWindow._registry.keys()):
            animation_tabs: list[FigureWidget] = []
            if not key in TabbedPlotWindow._registry:
                continue  # in case window was closed during iteration
            window = TabbedPlotWindow._registry[key]
            for tabs in window.tab_groups:
                for tab_id in tabs._figure_widgets:
                    if tabs[tab_id]._callback_registered:
                        animation_tabs.append(tabs[tab_id])
            animations[window] = animation_tabs

        if len(animations) == 0:
            QtWidgets.QMessageBox.information(
                None,
                "No Animations Found!",
                "No animation callbacks have been registered.",
            )
            return

        save_dialog = QtWidgets.QDialog()
        save_dialog.setWindowTitle("Save Animations")
        save_dialog.setMaximumHeight(800)
        save_dialog.setMinimumWidth(400)
        layout = QtWidgets.QVBoxLayout(save_dialog)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)

        scroll_area = QtWidgets.QScrollArea(widgetResizable=True)
        layout.addWidget(scroll_area)

        container = QtWidgets.QWidget()
        scroll_area.setWidget(container)
        container_layout = QtWidgets.QVBoxLayout(container)
        container_layout.setAlignment(QtCore.Qt.AlignTop)
        container_layout.setSpacing(0)
        container_layout.setContentsMargins(10, 10, 10, 10)

        bottom_row = QtWidgets.QWidget()
        bottom_layout = QtWidgets.QHBoxLayout(bottom_row)
        bottom_layout.setContentsMargins(0, 0, 0, 0)

        status_label = QtWidgets.QLabel("")
        bottom_layout.addWidget(status_label)

        bottom_layout.addStretch()
        cancel_button = QtWidgets.QPushButton("Close")
        cancel_button.clicked.connect(save_dialog.reject)
        bottom_layout.addWidget(cancel_button)

        for win in animations:
            window_label = QtWidgets.QLabel(f"Window: {win.id}")
            window_label.setStyleSheet("font-weight: bold; font-size: 14px;")
            container_layout.addWidget(window_label)

            for tab in animations[win]:
                row = QtWidgets.QWidget()
                row_layout = QtWidgets.QHBoxLayout(row)
                tab_label = QtWidgets.QLabel(f"    Tab: {tab._id}")
                tab_label.setStyleSheet("font-size: 12px;")
                row_layout.addWidget(tab_label)
                row_layout.addStretch()
                save_button = QtWidgets.QPushButton("Save")

                def make_save_fn(
                    current_tab: FigureWidget, button: QtWidgets.QPushButton
                ):
                    def save_fn():
                        status_label.setText(f"Saving...")
                        save_path = current_tab.save_animation(
                            frames, ts, save_dir=save_dir, parent=save_dialog, **kwargs
                        )
                        if save_path:
                            status_label.setText(f"Saved {save_path.name}")
                            button.setText("Saved!")
                            button.setEnabled(False)
                        else:
                            status_label.setText(f"Save canceled/failed")

                    return save_fn

                save_button.clicked.connect(make_save_fn(tab, save_button))
                row_layout.addWidget(save_button)
                container_layout.addWidget(row)
        layout.addWidget(bottom_row)

        save_dialog.adjustSize()
        save_dialog.exec()

    @staticmethod
    def set_animation_savedir(save_dir: str | Path | None) -> None:
        """
        Sets the default directory to save animations when using the animation
        player or the `save_animations` dialog.

        Args:
            save_dir: The directory to save the videos in. If None, will reset
                to the default save directory, which looks for the first valid
                path in this order: 1) "~/Videos" 2) "~/Documents" 3) "~".
        """
        FigureWidget.set_animation_savedir(save_dir)

    @staticmethod
    def close_all_windows() -> None:
        """
        Closes all created windows.
        """
        player = AnimationPlayer.instance()
        if player:
            player.close()
        for key in list(TabbedPlotWindow._registry.keys()):
            if not key in TabbedPlotWindow._registry:
                continue  # in case window was closed elsewhere during iteration
            window = TabbedPlotWindow._registry[key]
            window.qt.close()

    @staticmethod
    def get_screen_size() -> tuple[int, int]:
        """
        Returns the size of the screen in pixels. Tries to get the screen size
        of the screen at the current cursor position. If no screen is found,
        it will return the size of the primary screen.

        Returns:
            (width, height) (tuple[int,int]): The width and height of the screen
                in pixels.
        """
        assert isinstance(TabbedPlotWindow._app, QtWidgets.QApplication)
        screen = TabbedPlotWindow._app.screenAt(QtGui.QCursor.pos())
        if screen is None:
            # Fallback to primary screen if no screen is found at cursor position
            screen = TabbedPlotWindow._app.primaryScreen()
        size = screen.size()
        return size.width(), size.height()



from matplotlib.backends.qt_compat import QtWidgets
from typing import Callable

from .animation_player import AnimationPlayer


class CustomWidget(QtWidgets.QWidget):
    """
    A Qt widget that contains a custom Qt widget. Inherits from `QWidget`. This
    is used solely to provide a way to register animation callbacks for custom
    widgets so that they can be updated during animations.

    Methods:
        `update_widget`: Updates the widget with the registered callback function.
        `register_animation_callback`: Registers a callback function for how to
            update the widget during an animation.
    """

    def __init__(
        self, widget: QtWidgets.QWidget, add_animation_player: bool = False, parent=None
    ):
        """
        Initializes the CustomWidget.

        Args:
            widget: The custom Qt widget to include.
            add_animation_player: Whether to include an animation player widget
                in this tab (play, pause, etc.). Only works if animation callbacks
                are registered.
            parent: The parent widget for this widget.
        """
        super().__init__(parent)
        layout = QtWidgets.QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widget, stretch=1)
        if add_animation_player:
            animation_player = AnimationPlayer(parent=self)
            layout.addWidget(animation_player, stretch=0)
        self.setLayout(layout)

        def callback(idx: int = 0) -> None:
            return

        self._animation_callback = callback
        self._callback_registered = False
        self._latest_callback_idx = 0

    def update_widget(self, callback_idx: int = 0) -> None:
        """
        Updates the custom widget during an animation by calling the registered
        callback function.

        Args:
            callback_idx: An index passed to the registered animation callback
                function. This index is intended to specify which frame in the
                animation to draw.
        """
        # Attempting to detect if the same frame as last time to avoid re-drawing
        if self._callback_registered and callback_idx == self._latest_callback_idx:
            # print("Skipping custom widget update; same frame as last time.")
            return
        self._animation_callback(callback_idx)
        self._latest_callback_idx = callback_idx

    def register_animation_callback(self, callback: Callable[[int], None]) -> None:
        """
        Registers a callback function for how to update the custom widget during
        an animation.

        Args:
            callback: A function specifying how to update the widget. The function
                should take a single integer argument, which is the index of the
                current frame in the animation to draw. Registering callbacks
                allows abracatabra to better manage the timing of updates.
        """
        self._animation_callback = callback
        self._callback_registered = True



from typing import Callable, Optional, Self
from matplotlib.backends.qt_compat import QtWidgets, QtCore, QtGui

from . import keys

# from PySide6 import QtWidgets, QtCore, QtGui


class AnimationPlayer(QtWidgets.QWidget):
    """
    A simple player widget for controlling animations. This is a singleton class,
    meaning that only one instance can exist at a time. It provides basic media
    player controls (play/pause, step forward/backward, jump to start/end). It
    controls all windows within the application that have registered animation
    callbacks.

    Methods:
        `setup`: Sets up the animation player with the given parameters.
        `step_frame`: Steps the animation forward by one step if not paused.
    Static Methods:
        `instance`: Returns the singleton instance of the AnimationPlayer.
    """

    _instance: Optional[Self] = None
    help_text = """Animation Player Controls:
    k or Space: Play/Pause
    Home: Restart
    End: Go to last frame
    LeftArrow: Step back 1 frame
    RightArrow: Step forward 1 frame
    Ctrl+LeftArrow: Jump back
    Ctrl+RightArrow: Jump forward

    Media_Play/Pause: Play/Pause
    Media_Previous: Jump back
    Media_Next: Jump forward
    Ctrl+Media_Previous: Step back 1 frame
    Ctrl+Media_Next: Step forward 1 frame
    """
    shortcuts: dict[str, list[QtGui.QKeySequence]] = {
        "toggle_play_pause": [
            QtGui.QKeySequence(keys.Key_K),
            QtGui.QKeySequence(keys.Key_Space),
            QtGui.QKeySequence(keys.Key_MediaTogglePlayPause),
            QtGui.QKeySequence(keys.Key_MediaPlay),
            QtGui.QKeySequence(keys.Key_MediaPause),
        ],
        "restart": [QtGui.QKeySequence(keys.Key_Home)],
        "go_to_end": [QtGui.QKeySequence(keys.Key_End)],
        "step_backward": [
            QtGui.QKeySequence(keys.Key_Left),
            QtGui.QKeySequence(keys.ControlModifier | keys.Key_MediaPrevious),
        ],
        "jump_backward": [
            QtGui.QKeySequence(keys.ControlModifier | keys.Key_Left),
            QtGui.QKeySequence(keys.Key_MediaPrevious),
        ],
        "step_forward": [
            QtGui.QKeySequence(keys.Key_Right),
            QtGui.QKeySequence(keys.ControlModifier | keys.Key_MediaNext),
        ],
        "jump_forward": [
            QtGui.QKeySequence(keys.ControlModifier | keys.Key_Right),
            QtGui.QKeySequence(keys.Key_MediaNext),
        ],
        "save_animation": [QtGui.QKeySequence(keys.ControlModifier | keys.Key_S)],
    }

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        """
        Initializes the AnimationPlayer widget.
        """
        if AnimationPlayer.instance():
            raise RuntimeError("Only one instance of AnimationPlayer is allowed.")
        AnimationPlayer._instance = self
        super().__init__(parent)

        self.paused = True
        self.current_frame = 0

        # populate with default values; will be updated in setup()
        self.end_frame = 0
        self.ts = 0.0
        s = str(self.ts)
        self.t_decimals = len(s.split(".")[-1]) if "." in s else 2
        end_time = self.end_frame * self.ts
        self.t_digits = len(f"{end_time:.{self.t_decimals}f}")
        self.end_time = f"{end_time:>{self.t_digits}.{self.t_decimals}f}"

        self.step = 10
        self.jump = 10
        self.update_callback = lambda i: None

        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)
        self.std_icon = self.style().standardIcon

        # Visual indicator for focused player (thin accent bar)
        self._focused = False
        self._focus_indicator = QtWidgets.QWidget(self)
        self._focus_indicator.setFixedHeight(3)
        self._focus_indicator.setStyleSheet("background-color: #2196F3;")
        self._focus_indicator.hide()
        self._update_focus_indicator_position()

        # Connect to global focus change signal to show/hide focus indicator
        app = QtWidgets.QApplication.instance()
        if app:
            app.focusChanged.connect(self._on_focus_changed)

        # buttons
        self.play_button = QtWidgets.QPushButton()
        self.play_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        icon = self.std_icon(QtWidgets.QStyle.StandardPixmap.SP_MediaPlay)
        self.play_button.setIcon(icon)
        self.play_button.setToolTip("Play (k/Space)")
        self.play_button.clicked.connect(self._on_play_clicked)

        self.restart_button = QtWidgets.QPushButton()
        self.restart_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        icon = self.std_icon(QtWidgets.QStyle.StandardPixmap.SP_MediaSkipBackward)
        self.restart_button.setIcon(icon)
        self.restart_button.setToolTip("Restart animation (Home)")
        self.restart_button.clicked.connect(self._on_restart_clicked)

        self.end_button = QtWidgets.QPushButton()
        self.end_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        icon = self.std_icon(QtWidgets.QStyle.StandardPixmap.SP_MediaSkipForward)
        self.end_button.setIcon(icon)
        self.end_button.setToolTip("Go to end of animation (End)")
        self.end_button.clicked.connect(self._on_end_clicked)

        self.prev_button = QtWidgets.QPushButton()
        self.prev_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        icon = self.std_icon(QtWidgets.QStyle.StandardPixmap.SP_ArrowBack)
        self.prev_button.setIcon(icon)
        self.prev_button.setToolTip("Step back 1 frame (LeftArrow)")
        self.prev_button.clicked.connect(self._on_prev_clicked)
        self.prev_button.setAutoRepeat(True)
        self.prev_button.setAutoRepeatDelay(500)
        self.prev_button.setAutoRepeatInterval(50)

        self.jump_back_button = QtWidgets.QPushButton()
        self.jump_back_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        icon = self.std_icon(QtWidgets.QStyle.StandardPixmap.SP_MediaSeekBackward)
        self.jump_back_button.setIcon(icon)
        self.jump_back_button.setToolTip("Jump back 0 frames (Ctrl+LeftArrow)")
        self.jump_back_button.clicked.connect(self._on_jump_back_clicked)
        self.jump_back_button.setAutoRepeat(self.prev_button.autoRepeat())
        self.jump_back_button.setAutoRepeatDelay(self.prev_button.autoRepeatDelay())
        self.jump_back_button.setAutoRepeatInterval(
            self.prev_button.autoRepeatInterval() * 2
        )

        self.next_button = QtWidgets.QPushButton()
        self.next_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        icon = self.std_icon(QtWidgets.QStyle.StandardPixmap.SP_ArrowForward)
        self.next_button.setIcon(icon)
        self.next_button.setToolTip("Step forward 1 frame (RightArrow)")
        self.next_button.clicked.connect(self._on_next_clicked)
        self.next_button.setAutoRepeat(self.prev_button.autoRepeat())
        self.next_button.setAutoRepeatDelay(self.prev_button.autoRepeatDelay())
        self.next_button.setAutoRepeatInterval(self.prev_button.autoRepeatInterval())

        self.jump_forward_button = QtWidgets.QPushButton()
        self.jump_forward_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        icon = self.std_icon(QtWidgets.QStyle.StandardPixmap.SP_MediaSeekForward)
        self.jump_forward_button.setIcon(icon)
        self.jump_forward_button.setToolTip("Jump forward 0 frames (Ctrl+RightArrow)")
        self.jump_forward_button.clicked.connect(self._on_jump_forward_clicked)
        self.jump_forward_button.setAutoRepeat(self.prev_button.autoRepeat())
        self.jump_forward_button.setAutoRepeatDelay(
            self.jump_back_button.autoRepeatDelay()
        )
        self.jump_forward_button.setAutoRepeatInterval(
            self.jump_back_button.autoRepeatInterval()
        )

        self.save_button = QtWidgets.QPushButton()
        self.save_button.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        icon = self.std_icon(QtWidgets.QStyle.StandardPixmap.SP_DialogSaveButton)
        self.save_button.setIcon(icon)
        self.save_button.setToolTip("Save animation videos (Ctrl+s)")

        # slider
        self.slider = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal)
        self.slider.setMinimum(0)
        self.slider.setMaximum(0)
        self.slider.setValue(0)
        self.slider.valueChanged.connect(self._on_slider_changed)
        # self.slider.setTracking(False)  # only update on release, not while dragging
        self.slider.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)

        # spin box
        self.spin_box = QtWidgets.QSpinBox()
        self.spin_box.setMinimum(0)
        self.spin_box.setMaximum(0)
        self.spin_box.setValue(0)
        self.spin_box.setKeyboardTracking(False)
        self.spin_box.valueChanged.connect(self._on_spinbox_changed)
        self.spin_box.editingFinished.connect(self._on_spinbox_changed)

        # labels
        frame_label_1 = QtWidgets.QLabel("Frame:")
        self.frame_label_2 = QtWidgets.QLabel("/ 0")

        self.time_label = QtWidgets.QLabel()
        self._set_time_label()

        # layouts
        top_row = QtWidgets.QHBoxLayout()
        top_row.addWidget(self.jump_back_button)
        top_row.addWidget(self.prev_button)
        top_row.addWidget(self.slider)
        top_row.addWidget(self.next_button)
        top_row.addWidget(self.jump_forward_button)
        main_layout.addLayout(top_row)

        bottom_row = QtWidgets.QHBoxLayout()
        bottom_row.addWidget(self.play_button)
        bottom_row.addWidget(self.restart_button)
        bottom_row.addWidget(self.end_button)
        bottom_row.addWidget(self.save_button)
        bottom_row.addItem(
            QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Policy.Expanding)
        )
        bottom_row.addWidget(frame_label_1)
        bottom_row.addWidget(self.spin_box)
        bottom_row.addWidget(self.frame_label_2)
        bottom_row.addItem(
            QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Policy.Expanding)
        )
        bottom_row.addWidget(self.time_label)
        main_layout.addLayout(bottom_row)

        # create new window if no parent
        self._parent = parent
        if not parent:
            self.setWindowTitle("Animation Player")
            self.setMinimumWidth(600)
            play_icon = self.style().standardIcon(
                QtWidgets.QStyle.StandardPixmap.SP_MediaPlay
            )
            self.setWindowIcon(play_icon)
            self.show()
            self.raise_()
            self.adjustSize()
            self.quit_shortcuts = [
                QtGui.QShortcut(QtGui.QKeySequence(keys.Key_Q), self),
                QtGui.QShortcut(
                    QtGui.QKeySequence(keys.Key_Q | keys.ControlModifier), self
                ),
            ]
            for sc in self.quit_shortcuts:
                sc.activated.connect(self.close)
            self.register_shortcuts()

        # allow focus from clicks and tabbing
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        """
        Overrides the closeEvent to reset the singleton instance.

        Args:
            event (QCloseEvent): The close event.
        """
        super().closeEvent(event)
        AnimationPlayer._instance = None

    def _on_focus_changed(self, old_widget, new_widget):
        """
        Slot called when application focus changes. Updates the focus
        indicator based on whether the new focused widget is this player
        or one of its descendants.

        Args:
            old_widget: The previously focused widget (may be None).
            new_widget: The newly focused widget (may be None).
        """
        # Check if the newly focused widget is this player or a child of it
        has_focus = (new_widget is not None and
                     (new_widget is self or self.isAncestorOf(new_widget)))

        if self._focused != has_focus:
            self._focused = has_focus
            self._update_focus_style()

    def _update_focus_indicator_position(self):
        """
        Position the focus indicator bar at the top edge of the widget.
        """
        self._focus_indicator.setGeometry(0, 0, self.width(), 3)
        self._focus_indicator.raise_()

    def resizeEvent(self, event):
        """
        Handle resize events to reposition the focus indicator bar.

        Args:
            event: The resize event.
        """
        super().resizeEvent(event)
        self._update_focus_indicator_position()

    def _update_focus_style(self):
        """
        Update the visual style of the player based on focus state.
        Shows a thin blue accent bar at the top when focused.
        """
        if self._focused:
            self._focus_indicator.show()
        else:
            self._focus_indicator.hide()

    def setup(
        self,
        frames: int,
        ts: float,
        step: int,
        update_callback: Optional[Callable[[int], None]] = None,
    ) -> None:
        """
        Sets up the animation player with the given parameters. This essentially
        initializes the player, but allows it to be created in a tab beforehand.

        Args:
            frames (int): The total number of frames in the animation.
            ts (float): The time step between frames in seconds.
            step (int): How many frames to skip between each update.
            update_callback (Callable[[int], None] | None): A callback function
                that is called whenever the frame is changed. The function should
                take a single integer argument, which is the current frame index.
        """
        self.end_frame = frames - 1
        self.ts = ts
        s = str(ts)
        self.t_decimals = len(s.split(".")[-1]) if "." in s else 2

        end_time = self.end_frame * self.ts
        self.t_digits = len(f"{end_time:.{self.t_decimals}f}")
        self.end_time = f"{end_time:>{self.t_digits}.{self.t_decimals}f}"

        self.step = step
        self.jump = int(frames // 20)
        self.update_callback = update_callback or (lambda i: None)

        # relevant UI elements
        self.slider.setMaximum(frames - 1)
        self.spin_box.setMaximum(frames - 1)
        self.jump_back_button.setToolTip(
            f"Jump back {self.jump} frames (Ctrl+LeftArrow)"
        )
        self.jump_forward_button.setToolTip(
            f"Jump forward {self.jump} frames (Ctrl+RightArrow)"
        )
        self.slider.setPageStep(self.jump)
        self.frame_label_2.setText(f"/ {self.end_frame}")
        self._set_time_label()

    def step_frame(self) -> bool:
        """
        Steps the animation forward by one step if not paused.

        Returns:
            stepped (bool): True if the frame was stepped forward, False if
                the animation is paused.
        """
        if not self.paused:
            self.current_frame += self.step
            if self.current_frame > self.end_frame:
                self.current_frame = self.end_frame
                self._pause()
            self.slider.setValue(self.current_frame)
            self.update_callback(self.current_frame)
            return True
        return False

    def register_shortcuts(self, window: QtWidgets.QWidget = None) -> None:
        """
        Registers shortcuts for the animation player on the given window.

        Args:
            window (QWidget | None): The window to register the shortcuts on. If None,
                shortcuts will be registered on the animation player itself.
        """
        for action, shortcuts in self.shortcuts.items():
            for shortcut in shortcuts:
                sc = QtGui.QShortcut(shortcut, window or self)
                sc.setContext(QtCore.Qt.ShortcutContext.WindowShortcut)
                match action:
                    case "toggle_play_pause":
                        sc.activated.connect(self.play_button.click)
                    case "step_forward":
                        sc.activated.connect(self.next_button.click)
                    case "step_backward":
                        sc.activated.connect(self.prev_button.click)
                    case "jump_forward":
                        sc.activated.connect(self.jump_forward_button.click)
                    case "jump_backward":
                        sc.activated.connect(self.jump_back_button.click)
                    case "restart":
                        sc.activated.connect(self.restart_button.click)
                    case "go_to_end":
                        sc.activated.connect(self.end_button.click)
                    case "save_animation":
                        sc.activated.connect(self.save_button.click)
                    case _:
                        raise ValueError(
                            f"Unknown action '{action}' for shortcut '{shortcut}'"
                        )

    def _on_play_clicked(self):
        if self.paused and self.current_frame == self.end_frame:
            self.restart_button.click()
        if self.paused:
            self._play()
        else:
            self._pause()

    def _on_restart_clicked(self):
        self.current_frame = 0
        self.slider.setValue(self.current_frame)
        self.update_callback(self.current_frame)

    def _on_end_clicked(self):
        self.current_frame = self.end_frame
        self.slider.setValue(self.current_frame)
        self.update_callback(self.current_frame)

    def _on_prev_clicked(self):
        if not self.paused:
            return
        if self.current_frame > 0:
            self.current_frame -= 1
            self.slider.setValue(self.current_frame)

    def _on_jump_back_clicked(self):
        if not self.paused:
            return
        self.current_frame -= self.jump
        if self.current_frame < 0:
            self.current_frame = 0
        self.slider.setValue(self.current_frame)
        self.update_callback(self.current_frame)

    def _on_next_clicked(self):
        if not self.paused:
            return
        if self.current_frame < self.end_frame:
            self.current_frame += 1
            self.slider.setValue(self.current_frame)
            self.update_callback(self.current_frame)

    def _on_jump_forward_clicked(self):
        if not self.paused:
            return
        self.current_frame += self.jump
        if self.current_frame > self.end_frame:
            self.current_frame = self.end_frame
        self.slider.setValue(self.current_frame)
        self.update_callback(self.current_frame)

    def _on_slider_changed(self, value: int):
        self.current_frame = value
        self.spin_box.setValue(self.current_frame)
        self._set_time_label()
        self.update_callback(self.current_frame)

    def _on_spinbox_changed(self):
        value = self.spin_box.value()
        self.current_frame = value
        self.slider.setValue(self.current_frame)

    def _set_time_label(self):
        time = self.current_frame * self.ts
        time = f"{time:>{self.t_digits}.{self.t_decimals}f}"
        self.time_label.setText(f"Sim Time: {time} / {self.end_time} s")

    def _pause(self):
        self.paused = True
        icon = self.std_icon(QtWidgets.QStyle.StandardPixmap.SP_MediaPlay)
        self.play_button.setIcon(icon)
        self.play_button.setToolTip("Play")
        self.jump_back_button.setEnabled(True)
        self.prev_button.setEnabled(True)
        self.next_button.setEnabled(True)
        self.jump_forward_button.setEnabled(True)

    def _play(self):
        self.paused = False
        icon = self.std_icon(QtWidgets.QStyle.StandardPixmap.SP_MediaPause)
        self.play_button.setIcon(icon)
        self.play_button.setToolTip("Pause")
        self.jump_back_button.setEnabled(False)
        self.prev_button.setEnabled(False)
        self.next_button.setEnabled(False)
        self.jump_forward_button.setEnabled(False)

    @classmethod
    def instance(cls) -> Optional[Self]:
        """
        Returns:
            instance (AnimationPlayer | None): The singleton instance of the
                AnimationPlayer, or None if it does not exist.
        """
        return cls._instance


from matplotlib.backends.qt_compat import QtWidgets, QtGui, QtCore
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt import NavigationToolbar2QT as NavigationToolbar
from typing import Callable
from pathlib import Path

from .animation_player import AnimationPlayer
from . import keys

# from PySide6 import QtWidgets, QtGui, QtCore


class CustomNavToolbar(NavigationToolbar):
    toolitems = []
    _shortcut_map = {
        "Home": "h",
        "Back": "c",
        "Forward": "v",
        "Pan": "p",
        "Zoom": "z",
        "Save": "s",
    }
    for text, tooltip, img, callback in NavigationToolbar.toolitems:
        if text in _shortcut_map:
            tooltip = f"{tooltip} ({_shortcut_map[text]})"
        toolitems.append((text, tooltip, img, callback))


class FigureWidget(QtWidgets.QWidget):
    """
    A Qt widget that contains a matplotlib figure canvas with an optional toolbar.
    Inherits from `QWidget`.

    Methods:
        `update_figure`: Updates the figure canvas if anything has changed.
        `show_toolbar`: Show or hide the navigation toolbar.
        `register_animation_callback`: Registers a callback function for how to
            update the figure during an animation.
    """

    help_text = """Figure Controls:
    h: Home (reset view)
    c: Back 1 view
    v: Forward 1 view
    p: Toggle pan mode
    z: Toggle zoom mode
    s: Save figure
    """
    _animation_savedir = Path.home() / "Videos"

    def __init__(
        self,
        name: str | int = "figure",
        blit: bool = False,
        include_toolbar: bool = True,
        add_animation_player: bool = False,
        parent=None,
    ):
        """
        Initializes the FigureWidget. This creates a matplotlib figure canvas
        and optionally includes a navigation toolbar.

        Args:
            name: The name of the figure widget, used as the default filename
                when saving the figure.
            blit: If True, enables blitting for faster rendering.
            include_toolbar: If True, includes a navigation toolbar with the canvas.
            add_animation_player: Whether to include an animation player widget
                in this tab (play, pause, etc.). Only works if animation callbacks
                are registered.
            parent: The parent widget for this widget.
        """
        super().__init__(parent)
        self.blit = blit
        layout = QtWidgets.QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)

        self.canvas = FigureCanvas()
        # override default save behavior to use pdf and custom filename
        if isinstance(name, int):
            name = f"figure_{name}"
        self._id: str = name
        name = name.replace(" ", "_")
        self.canvas.get_default_filetype = lambda: "pdf"
        self.canvas.get_default_filename = lambda: f"{name}.pdf"

        self.figure = self.canvas.figure
        # self.figure.set_layout_engine('tight') # slows down rendering ~2x
        # self.figure.tight_layout() # does not seem to do anything here

        self.toolbar = CustomNavToolbar(self.canvas, self)
        self.toolbar.setMaximumHeight(25)
        self.toolbar.setVisible(include_toolbar)

        layout.addWidget(self.canvas, stretch=1)
        layout.addWidget(self.toolbar)

        if add_animation_player:
            animation_player = AnimationPlayer(parent=self)
            layout.addWidget(animation_player, stretch=0)

        self.setLayout(layout)

        self._update_callback: Callable[[int], None] = lambda i: None
        self._callback_registered = False
        self._latest_callback_idx = 0

        self._setup_shortcuts()

    def update_figure(self, callback_idx: int = 0) -> None:
        """
        Updates the figure canvas if anything has changed. If blitting is
        enabled, it will only redraw the parts of the figure that have changed.
        If not, it will redraw the entire canvas. NOTE that blitting requires
        the user to manage the background and artist updates manually, i.e., the
        user must call `canvas.copy_from_bbox()` and `canvas.restore_region()`
        at the appropriate times AND ensure that the artists are drawn before
        calling this method.

        Args:
            callback_idx: An index passed to the registered animation callback
                function. This index is intended to specify which frame of the
                animation to draw, if an animation callback has been registered.
        """
        # Attempting to detect if the same frame as last time to avoid re-drawing
        if self._callback_registered and callback_idx == self._latest_callback_idx:
            # print("Skipping figure update; same frame as last time.")
            return
        self._update_callback(callback_idx)
        self._latest_callback_idx = callback_idx
        if not self.figure.stale:
            return
        if self.blit:
            self.canvas.blit()
        else:
            self.canvas.draw_idle()
        self.canvas.flush_events()

    def show_toolbar(self, show: bool = True) -> None:
        """
        Show or hide the navigation toolbar.

        Args:
            show: If True, shows the toolbar. If False, hides it.
        """
        self.toolbar.setVisible(show)

    def register_animation_callback(self, callback: Callable[[int], None]) -> None:
        """
        Registers a callback function for how to update the figure during an
        animation. Note that if the figure has multiple axes or artists, the
        user is responsible for managing the updates to all of those objects in
        the callback function (callback is per figure not per axis/artist).

        Args:
            callback: A function specifying how to update the widget. The function
                should take a single integer argument, which is the index of the
                current frame in the animation to draw. Registering callbacks
                allows abracatabra to better manage the timing of updates.
        """
        self._update_callback = callback
        self._callback_registered = True

    def save_animation(
        self,
        frames: int,
        dt: float | None = None,
        filename: str | None = None,
        save_dir: str | Path | None = None,
        parent: QtWidgets.QWidget | None = None,
        **kwargs,
    ) -> Path | None:
        """
        Saves the animation for this figure if an animation callback has been
        registered. This is a convenience function that calls the registered
        animation callback for each frame and saves the resulting frames as an
        MP4 video (recommended) or gif.

        Args:
            frames: The number of frames in the animation.
            dt (seconds): The time step between frames in seconds. If None,
                matplotlib will use default interval.
            filename: The filename to save the animation to. If None, uses the
                default filename specified by `canvas.get_default_filename()`.
            **kwargs: Additional keyword arguments passed to
                `matplotlib.animation.FuncAnimation.save()`.

        Returns:
            path: Path to the saved file, or None if the save dialog was cancelled.
        """
        from matplotlib.animation import FuncAnimation

        if not self._callback_registered:
            raise RuntimeError("No animation callback registered for this figure.")
        if save_dir:
            self.set_animation_savedir(save_dir)
        video_dir = self.get_animation_savedir()

        if filename:
            save_file = video_dir / filename
        else:
            default_filename = self.canvas.get_default_filename()
            default_filename = Path(default_filename).with_suffix(".mp4")
            save_file, _ = QtWidgets.QFileDialog.getSaveFileName(
                parent or self,
                "Save Animation",
                str(video_dir / default_filename),
                "MP4 Video (*.mp4);;GIF Image (*.gif);;All Files (*)",
            )
        if not save_file:
            return None
        savedir = Path(save_file).parent
        if savedir.exists():
            self.set_animation_savedir(savedir)

        def animate(idx):
            self._update_callback(idx)
            return ()

        anim = FuncAnimation(
            self.figure,
            animate,
            frames,
            interval=dt * 1000 if dt else None,
            blit=self.blit,
        )
        path = Path(save_file)
        if not path.suffix in [".mp4", ".gif"]:
            path = path.with_suffix(".mp4")  # default to mp4 if no valid extension

        progress = QtWidgets.QProgressDialog(
            f"Destination: {path}", "Cancel", 0, frames, parent or self
        )
        progress.setWindowTitle("Saving Animation...")
        progress.setWindowModality(QtCore.Qt.ApplicationModal)
        progress.setCancelButton(None)  # Removes the cancel button
        progress.setMinimumDuration(0)

        def progress_callback(frame_idx, total_frames):
            progress.setValue(frame_idx)
            if progress.wasCanceled():
                anim.event_source.stop()

        anim.save(path, progress_callback=progress_callback, **kwargs)
        progress.close()
        return path

    def _setup_shortcuts(self):
        shortcuts = {
            keys.Key_P: self.toolbar.pan,
            keys.Key_H: self.toolbar.home,
            keys.Key_Z: self.toolbar.zoom,
            keys.Key_C: self.toolbar.back,
            keys.Key_V: self.toolbar.forward,
            keys.Key_S: self.toolbar.save_figure,
        }
        for key, callback in shortcuts.items():
            shortcut = QtGui.QShortcut(QtGui.QKeySequence(key), self)
            shortcut.setContext(QtCore.Qt.WidgetWithChildrenShortcut)
            shortcut.activated.connect(callback)

    @staticmethod
    def set_animation_savedir(path: str | Path | None = None) -> None:
        """
        Sets the default directory for saving animations.

        Args:
            path: The path to the directory where animations should be saved by
                default. If None, uses the user's Videos directory if it exists,
                otherwise falls back to Documents, then home directory.
        """
        if path:
            save_dir = Path(path)
        else:
            save_dir = Path.home() / "Videos"

        if not save_dir.exists():
            save_dir = Path.home() / "Documents"
        if not save_dir.exists():
            save_dir = Path.home()
        FigureWidget._animation_savedir = save_dir

    @staticmethod
    def get_animation_savedir() -> Path:
        """
        Gets the current default directory for saving animations.

        Returns:
            The path to the directory where animations are currently saved by default.
        """
        return FigureWidget._animation_savedir



import numpy as np
import abracatabra


def test_readme_example():
    window1 = abracatabra.TabbedPlotWindow(window_id="test", ncols=2)
    window2 = abracatabra.TabbedPlotWindow(size=(500, 400))

    # data
    t = np.arange(0, 10, 0.001)
    ysin = np.sin(t)
    ycos = np.cos(t)

    fig = window1.add_figure_tab("sin", col=0)
    ax = fig.add_subplot()
    (line1,) = ax.plot(t, ysin, "--")
    ax.set_xlabel("time")
    ax.set_ylabel("sin(t)")
    ax.set_title("Plot of sin(t)")

    fig = window1.add_figure_tab("time", col=1)
    ax = fig.add_subplot()
    ax.plot(t, t)
    ax.set_xlabel("time")
    ax.set_ylabel("t")
    ax.set_title("Plot of t")

    window1.apply_tight_layout()

    fig = window2.add_figure_tab("cos")
    ax = fig.add_subplot()
    (line2,) = ax.plot(t, ycos, "--")
    ax.set_xlabel("time")
    ax.set_ylabel("cos(t)")
    ax.set_title("Plot of cos(t)")

    fig = window2.add_figure_tab("sin^2")
    fig_test = window2.add_figure_tab("sin^2")
    assert fig is fig_test
    ax = fig.add_subplot()
    ax.plot(t, ysin**2)
    ax.set_xlabel("time")
    ax.set_ylabel("t")
    ax.set_title("Plot of t", fontsize=20)

    window2.apply_tight_layout()

    # animate
    dt = 0.1
    for k in range(100):
        t += dt
        ysin = np.sin(t)
        line1.set_ydata(ysin)
        ycos = np.cos(t)
        line2.set_ydata(ycos)
        abracatabra.update_all_windows(0.01)

    abracatabra.abracatabra(block=False)
    assert True


if __name__ == "__main__":
    test_readme_example()



from typing import Sequence
from itertools import chain

from .tabbed_figure_widget import TabbedFigureWidget


class TabGroupContainer:
    def __init__(self, group_list: list[list[TabbedFigureWidget]], row_major: bool):
        """
        A container for a list of tab groups than can handle non-homogeneous
        layouts and indexing, i.e., elements of the inner list do not need to
        have the same length. Must index with two indices, even if there is
        only one item in the list.

        Args:
            group_list: A list of lists, where each inner list contains
                TabbedFigureWidget objects representing a group of tabs.
            row_major: If True, the first index is the row index and the second
                index is the column index. If False, the order is reversed.
        """
        self._tab_groups: list[list[TabbedFigureWidget]] = group_list
        self._row_major = row_major

    def __getitem__(self, index: tuple[int, int]) -> TabbedFigureWidget:
        """
        Returns the tab group at the given index.

        Args:
            index: A tuple containing two integers, where the first integer is
                the row index and the second integer is the column index. Must
                use 2 indices, even if there is only one item.
        """
        if not isinstance(index, Sequence) or len(index) != 2:
            raise ValueError(f"Index must contain two integers, got {index}")
        row, col = index
        if not self._row_major:
            col, row = index
        return self._tab_groups[row][col]

    def __iter__(self):
        """
        Iterates over every tab group in the container.
        """
        return chain.from_iterable(self._tab_groups)



import os
from matplotlib.figure import Figure
from matplotlib.backends.qt_compat import QtWidgets, QtCore, QtGui

from .figure_widget import FigureWidget
from .custom_widget import CustomWidget

# Suppress atspi accessibility warnings from Qt (started happening after using slots)
os.environ["QT_LOGGING_RULES"] = "qt.accessibility.atspi=false"


class TabbedFigureWidget(QtWidgets.QTabWidget):
    """
    A Qt widget that can contain multiple tabs, each with a matplotlib Figure
    or custom widget. This class inherits from QTabWidget to create a tabbed
    interface with visual focus indication (blue accent bar when focused).

    The widget automatically manages focus by transferring it to tab contents
    when tabs are clicked or changed, ensuring keyboard shortcuts work properly.
    An event filter tracks focus changes to show/hide the focus indicator.

    Methods:
        `update_active_tab`: Updates the currently active tab's widget.
        `add_figure_tab`: Adds a new tab with a matplotlib Figure.
        `add_custom_tab`: Adds a new tab with a custom Qt widget.
        `get_tab`: Returns the widget associated with a given tab ID.
        `set_tab_position`: Sets the position of the tab bar.
        `set_tab_fontsize`: Sets the font size of the tab bar.
    """

    def __init__(self, autohide: bool, position: str = "top", fontsize: int = 8):
        """
        Initializes the TabbedFigureWidget.

        Args:
            autohide (bool): If True, the tab bar will auto-hide when there is
                only one tab.
            position (str): The position of the tab bar. Can be 'top', 'bottom',
                'left', or 'right' as well as 'north', 'south', 'east', or
                'west' (only first character is checked).
            fontsize (int): The font size of the tab labels.
        """
        super().__init__()
        tabbar = self.tabBar()
        tabbar.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        assert isinstance(tabbar, QtWidgets.QTabBar)
        tabbar.setAutoHide(autohide)
        tabbar.setContentsMargins(0, 0, 0, 0)
        self.set_tab_position(position)
        self.set_tab_fontsize(fontsize)
        self._figure_widgets: dict[str, FigureWidget] = {}
        self._custom_widgets: dict[str, CustomWidget] = {}
        self._latest_callback_idx = 0
        self.currentChanged.connect(self._on_tab_changed)
        self.tabBarClicked.connect(self._on_tab_clicked)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

        # Visual indicator for focused tab group (thin accent bar)
        self._focused = False
        self._focus_indicator = QtWidgets.QWidget(self)
        self._focus_indicator.setFixedHeight(3)
        self._focus_indicator.setStyleSheet("background-color: #2196F3;")
        self._focus_indicator.hide()
        self._update_focus_indicator_position()
        self._update_focus_style()

        # Connect to global focus change signal instead of using event filters
        # This eliminates race conditions from deferred callbacks during cleanup
        app = QtWidgets.QApplication.instance()
        if app:
            app.focusChanged.connect(self._on_focus_changed)

    def __getitem__(self, tab_id: str | int) -> FigureWidget | CustomWidget:
        """
        Provides dictionary-like access to tabs by their ID for convenience.

        Args:
            tab_id (str|int): The title/ID of the tab.

        Returns:
            widget (FigureWidget | CustomWidget): The widget associated with the
                given tab ID.
        """
        return self.get_tab(tab_id)

    def _on_focus_changed(self, old_widget, new_widget):
        """
        Slot called when application focus changes. Updates the focus
        indicator based on whether the new focused widget is a descendant
        of this tab group.

        This approach is cleaner than using event filters and deferred
        callbacks, eliminating race conditions during widget cleanup.

        Args:
            old_widget: The previously focused widget (may be None).
            new_widget: The newly focused widget (may be None).
        """
        # Check if the newly focused widget is a child of this tab group
        has_focus = new_widget is not None and self.isAncestorOf(new_widget)

        if self._focused != has_focus:
            self._focused = has_focus
            self._update_focus_style()

    def _update_focus_indicator_position(self):
        """
        Position the focus indicator bar at the top edge of the widget.
        """
        self._focus_indicator.setGeometry(0, 0, self.width(), 3)
        self._focus_indicator.raise_()

    def resizeEvent(self, event):
        """
        Handle resize events to reposition the focus indicator bar.

        Args:
            event: The resize event.
        """
        super().resizeEvent(event)
        self._update_focus_indicator_position()

    def _update_focus_style(self):
        """
        Update the visual style of the tab group based on focus state.
        Shows a thin blue accent bar at the top when focused, works even
        with autohide tabs. No changes to tab bar to preserve original look.
        """
        if self._focused:
            # Show blue accent bar when focused
            self._focus_indicator.show()
        else:
            # Hide accent bar when not focused
            self._focus_indicator.hide()

    def update_active_tab(self, callback_idx: int = 0) -> None:
        """
        Updates the currently active tab's widget.

        Args:
            callback_idx (int): An index passed to the registered animation
                callback function. This index is intended to specify which frame
                to draw.
        """
        self._latest_callback_idx = callback_idx
        active_widget = self.currentWidget()
        if isinstance(active_widget, FigureWidget):
            active_widget.update_figure(callback_idx)
        elif isinstance(active_widget, CustomWidget):
            active_widget.update_widget(callback_idx)

    def add_figure_tab(
        self,
        tab_id: str | int,
        blit: bool = False,
        include_toolbar: bool = True,
        add_animation_player: bool = False,
    ) -> Figure:
        """
        Adds a new tab to the widget with the given title/tab_id, which
        creates and returns a matplotlib Figure. Tabs are displayed in the
        order they are added. If the tab ID already exists, returns the
        existing Figure without creating a new tab.

        The new tab widget has an event filter installed to track focus changes
        for the focus indicator.

        Args:
            tab_id (str|int): The title/ID of the tab. If the tab ID already
                exists, the existing Figure from that tab will be returned.
            blit (bool): If True, enables blitting for faster rendering on the
                Figure in this tab.
            include_toolbar (bool): If True, includes a navigation toolbar
                with the Figure in this tab.
            add_animation_player (bool): Whether to include an animation player
                widget in this tab (play, pause, etc.). Only works if animation
                callbacks are registered.

        Returns:
            figure: The matplotlib Figure object for the tab.
        """
        id_ = str(tab_id)
        if id_ in self._figure_widgets:
            return self._figure_widgets[id_].figure
        new_tab = FigureWidget(
            tab_id, blit, include_toolbar, add_animation_player, self
        )
        self._figure_widgets[id_] = new_tab
        idx = self.currentIndex()
        super().addTab(new_tab, id_)
        self.setCurrentWidget(new_tab)  # activate tab to auto size figure
        self.setCurrentIndex(idx)  # switch back to original tab
        return new_tab.figure

    def add_custom_tab(
        self,
        widget: QtWidgets.QWidget,
        tab_id: str | int,
        add_animation_player: bool = False,
    ) -> None:
        """
        Adds a new tab to the widget with the given title/tab_id, which
        contains the provided custom Qt widget. Tabs are displayed in the
        order they are added.

        The new tab widget has an event filter installed to track focus changes
        for the focus indicator.

        Args:
            widget (QWidget): The custom Qt widget to add as a tab.
            tab_id (str|int): The title/ID of the tab. Must be unique.
            add_animation_player (bool): Whether to include an animation player
                widget in this tab (play, pause, etc.). Only works if animation
                callbacks are registered.

        Raises:
            ValueError: If a tab with the given ID already exists.
        """
        id_ = str(tab_id)
        if id_ in self._figure_widgets | self._custom_widgets:
            raise ValueError(f"Tab with id '{id_}' already exists.")
        new_tab = CustomWidget(widget, add_animation_player)
        self._custom_widgets[id_] = new_tab
        super().addTab(new_tab, id_)
        return

    def get_tab(self, tab_id: str | int) -> FigureWidget | CustomWidget:
        """
        Returns the widget associated with the given tab ID.

        Args:
            tab_id (str|int): The title/ID of the tab.

        Returns:
            widget: The widget associated with the given tab ID.

        Raises:
            ValueError: If no tab with the given ID exists.
        """
        id_ = str(tab_id)
        if id_ in self._figure_widgets:
            return self._figure_widgets[id_]
        elif id_ in self._custom_widgets:
            return self._custom_widgets[id_]
        else:
            raise ValueError(f"Tab with id '{id_}' does not exist.")

    def set_tab_position(self, position: str = "top") -> None:
        """
        Sets the position of the tab bar.

        Args:
            position (str): The position of the tab bar. Can be 'top', 'bottom',
                'left', or 'right' as well as 'north', 'south', 'east', or
                'west' (only first character is checked).
        """
        char = position[0].lower()
        if char in ["b", "s"]:
            self.setTabPosition(QtWidgets.QTabWidget.TabPosition.South)
        elif char in ["l", "w"]:
            self.setTabPosition(QtWidgets.QTabWidget.TabPosition.West)
        elif char in ["r", "e"]:
            self.setTabPosition(QtWidgets.QTabWidget.TabPosition.East)
        else:
            self.setTabPosition(QtWidgets.QTabWidget.TabPosition.North)

    def set_tab_fontsize(self, fontsize: int) -> None:
        """
        Sets the font size of the tab bar.

        Args:
            fontsize (int): The font size to set for the tab bar.
        """
        tabbar = self.tabBar()
        assert isinstance(tabbar, QtWidgets.QTabBar)
        font = tabbar.font()
        font.setPointSize(fontsize)
        tabbar.setFont(font)

    @QtCore.Slot(int)
    def _on_tab_changed(self, index: int) -> None:
        """
        Slot called when the current tab is changed. This is used to make sure
        the animation callback is called for the newly active tab, and to
        transfer focus to the newly active widget so its shortcuts work.

        Args:
            index (int): The index of the newly selected tab.
        """
        # Transfer focus to current widget so its shortcuts work when tab is clicked
        current_widget = self.currentWidget()
        if current_widget:
            current_widget.setFocus()

        if self._latest_callback_idx > 0:
            self.update_active_tab(self._latest_callback_idx)

    @QtCore.Slot(int)
    def _on_tab_clicked(self, index: int) -> None:
        """
        Slot called when a tab is clicked. This transfers focus to the clicked
        tab's widget, even if it was already the active tab. This ensures
        shortcuts work when clicking on an already-visible tab.

        Args:
            index (int): The index of the clicked tab.
        """
        current_widget = self.currentWidget()
        if current_widget:
            current_widget.setFocus()



import numpy as np
import abracatabra
from matplotlib.animation import FuncAnimation


def test_animation(attached_player: bool, attach_location: str = "figure", save=False):
    blit = True
    window = abracatabra.TabbedPlotWindow(
        autohide_tabs=True,
        open_window=True,
        add_animation_player=attached_player and not attach_location == "figure",
    )

    fig1 = window.add_figure_tab(
        "robot arm animation",
        include_toolbar=True,
        blit=blit,
        add_animation_player=attached_player and attach_location == "figure",
    )

    # window.show()

    ax1 = fig1.add_subplot()

    # background elements
    fig1.tight_layout()
    ax1.set_aspect("equal", "box")
    length = 1.0
    lim = 1.25 * length
    ax1.axis((-lim, lim, -lim, lim))
    (baseline,) = ax1.plot([0, length], [0, 0], "k--")

    # draw and save background for fast rendering
    fig1.canvas.draw()
    background1 = fig1.canvas.copy_from_bbox(ax1.bbox)  # type: ignore

    # moving elements
    def get_arm_endpoints(theta):
        x = np.array([0, length * np.cos(theta)])
        y = np.array([0, length * np.sin(theta)])
        return x, y

    t_hist = np.linspace(0, 10, 501)
    theta_hist = np.sin(t_hist)
    x, y = get_arm_endpoints(theta_hist[0])
    (arm_line,) = ax1.plot(x, y, linewidth=5, color="blue")

    def update(frame_idx):
        theta = theta_hist[frame_idx]
        x, y = get_arm_endpoints(theta)
        arm_line.set_xdata(x)
        arm_line.set_ydata(y)

        if blit:
            fig1.canvas.restore_region(background1)  # type: ignore
            ax1.draw_artist(arm_line)

    window.register_animation_callback(update, "robot arm animation")

    window2 = abracatabra.TabbedPlotWindow(ncols=2, autohide_tabs=True)
    fig2 = window2.add_figure_tab("sin", blit=blit)
    ax2 = fig2.add_subplot()
    (sin_line,) = ax2.plot(t_hist, theta_hist)
    fig2.tight_layout()
    fig2.canvas.draw()
    sin_line.set_data([], [])
    fig2.canvas.draw()
    background2 = fig2.canvas.copy_from_bbox(ax2.bbox)  # type: ignore

    def update_sin(frame_idx):
        sin_line.set_data(t_hist[:frame_idx], theta_hist[:frame_idx])
        fig2.canvas.restore_region(background2)  # type: ignore
        ax2.draw_artist(sin_line)

    window2.register_animation_callback(update_sin, "sin")

    fig3 = window2.add_figure_tab("cos", blit=blit)
    ax3 = fig3.add_subplot()
    t_cos = t_hist - 5
    (cos_line,) = ax3.plot(t_cos, np.cos(t_cos))
    fig3.canvas.draw()
    cos_line.set_data([], [])
    fig3.canvas.draw()
    background3 = fig3.canvas.copy_from_bbox(ax3.bbox)  # type: ignore

    def update_cos(frame_idx):
        cos_line.set_data(t_cos[:frame_idx], np.cos(t_cos[:frame_idx]))
        fig3.canvas.restore_region(background3)  # type: ignore
        ax3.draw_artist(cos_line)

    window2.register_animation_callback(update_cos, "cos")

    fig4 = window2.add_figure_tab("wave", blit=blit, col=1)
    ax4 = fig4.add_subplot()
    (wave_line,) = ax4.plot(t_hist, theta_hist)
    fig4.canvas.draw()
    wave_line.set_data([], [])
    fig4.canvas.draw()
    background4 = fig4.canvas.copy_from_bbox(ax4.bbox)  # type: ignore

    def update_wave(frame_idx):
        wave_line.set_data(t_hist, np.sin(t_hist + frame_idx * 0.1))
        fig4.canvas.restore_region(background4)  # type: ignore
        ax4.draw_artist(wave_line)

    window2.register_animation_callback(update_wave, "wave", col=1)

    dt = 0.01
    # if save:
    #     tab = window.tab_groups[0, 0].get_tab("robot arm animation")
    #     tab.save_animation(frames=len(theta_hist), dt=dt, filename="test.mp4")

    abracatabra.animate_all_windows(len(theta_hist), ts=dt, use_player=True)
    assert True


if __name__ == "__main__":
    test_animation(attached_player=False, save=True)
    test_animation(attached_player=True, attach_location="figure")
    test_animation(attached_player=True, attach_location="window")



import numpy as np
import abracatabra


def test_readme_blit_example():
    blit = True
    window = abracatabra.TabbedPlotWindow(autohide_tabs=True)
    fig = window.add_figure_tab("robot arm animation", include_toolbar=False, blit=blit)
    ax = fig.add_subplot()

    # background elements
    fig.tight_layout()
    ax.set_aspect("equal", "box")
    length = 1.0
    lim = 1.25 * length
    ax.axis((-lim, lim, -lim, lim))
    (baseline,) = ax.plot([0, length], [0, 0], "k--")

    # draw and save background for fast rendering
    fig.canvas.draw()
    background = fig.canvas.copy_from_bbox(ax.bbox)  # type: ignore

    # moving elements
    def get_arm_endpoints(theta):
        x = np.array([0, length * np.cos(theta)])
        y = np.array([0, length * np.sin(theta)])
        return x, y

    theta_hist = np.sin(np.linspace(0, 10, 501))
    x, y = get_arm_endpoints(theta_hist[0])
    (arm_line,) = ax.plot(x, y, linewidth=5, color="blue")

    # animate
    for theta in theta_hist:
        x, y = get_arm_endpoints(theta)
        arm_line.set_xdata(x)
        arm_line.set_ydata(y)

        if blit:
            fig.canvas.restore_region(background)  # type: ignore
            ax.draw_artist(arm_line)

        abracatabra.update_all_windows(0.01)

    # keep window open
    abracatabra.abracatabra(block=False)
    assert True


if __name__ == "__main__":
    test_readme_blit_example()
